__all__ = ['prior_transform', '_uniform_ppf', '_loguniform_ppf',  '_norm_ppf']

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

from dynesty.utils import print_fn
import io
import numpy as np
from scipy.special import erfinv
import time


def prior_transform(uniform_hypercube, fixed_durations):
    """
    Prior transform over physical ALDERAAN basis {C0, C1, r, b, T14}...{q1,q2}

    Distributions are hard-coded to be:
        Normal on ephemeris perturbations {C0,C1}
        Log-uniform on radius ratio r and transit duration T14
        Uniform on impact parameter b
        Uniform on quadratic limb darkening coefficients {q1,q2}

    For motivation behind this prior-parameter choice see:
        Carter+ 2008 (2008ApJ...689..499C)
        Kipping 2013 (2013MNRAS.435.2152K)
        Gilbert, MacDougall, & Petigura 2022 (2022AJ....164...92G)
        MacDougal, Gilbert, & Petigura 2023 (2023AJ....166...61M)

    Assumes that transit durations are known with reasonable confidence
      * sampled T14 will be restricted between (0.01*fixed_duration, 3*fixed_duration)

    Parameters
    ----------
        uniform_hypercube : array-like, length 5 * N + 2
            list of parameters N * [C0, C1, r, b, T14] + [q1, q2]
        fixed_durations : array-like
            list of transit durations for N planets (used for setting prior limits)

    Returns
    -------
        transformed_hypercube : array_like, length 5 * N + 2
            transformed samples
    """
    npl = len(fixed_durations)

    u = np.array(uniform_hypercube)  # U(0,1) priors
    x = np.zeros_like(u)             # physical priors

    # 5 * npl + 2 parameters: [C0, C1, r, b, T14]...[q1, q2]
    for n in range(npl):
        x[5 * n + 0] = _norm_ppf(u[0 + n * 5], 0.0, 0.1)
        x[5 * n + 1] = _norm_ppf(u[1 + n * 5], 0.0, 0.1)
        x[5 * n + 2] = _loguniform_ppf(u[2 + n * 5], 1e-5, 0.99)
        x[5 * n + 3] = _uniform_ppf(u[3 + n * 5], 0.0, 1 + x[5 * n + 2])
        x[5 * n + 4] = _loguniform_ppf(u[4 + n * 5], 0.01*fixed_durations[n], 3 * fixed_durations[n])

    # limb darkening coefficients (see Kipping 2013)
    x[-2] = _uniform_ppf(u[-2], 0, 1)
    x[-1] = _uniform_ppf(u[-1], 0, 1)

    return x


def _uniform_ppf(u, a, b):
    """
    Transform from U(0,1) --> Uniform(a,b)

    Parameters
    ----------
        u : array-like
            samples from standard uniform distribution U(0,1)
        a : float
            lower bound on transformed distribution U(a,b)
        b : float
            upper bound on transformed distribution U(a,b)

    Returns
    -------
        x : array-like
            transformed samples from uniform distribution
    """
    return u * (b - a) + a


def _loguniform_ppf(u, a, b):
    """
    Transform from U(0,1) --> LogUniform(a,b)

    Parameters
    ----------
        u : array-like
            samples from standard uniform distribution U(0,1)
        a : float
            lower bound on transformed distribution lnU(a,b)
        b : float
            upper bound on transformed distribution lnU(a,b)

    Returns
    -------
        x : array-like
            transformed samples from log-uniform distribution
    """
    return np.exp(u * np.log(b) + (1 - u) * np.log(a))


def _norm_ppf(u, mu, sig, eps=1e-12):
    """
    Transform from U(0,1) --> Normal(mu,sig)

    Parameters
    ----------
        u : array-like
            samples from standard uniform distribution U(0,1)
        mu : float
            mean of transformed distribution N(mu,sig)
        sig : float
            standard deviation of transformed distribution N(mu,sig)
        eps : float (optional)
            provides numerical stability in the inverse error function at edges

    Returns
    -------
        x : array-like
            transformed samples from Normal distribution
    """
    return mu + sig * np.sqrt(2) * erfinv((2 * u - 1) / (1 + eps))


def throttled_print_fn(interval=10):
    last = [0]
    start = time.time()

    def callback(results, niter, ncall, *args, **kwargs):
        now = time.time()
        if now - last[0] >= interval:
            runtime = (now - start)/60

            buf = io.StringIO()
            stdout = sys.stdout
            sys.stdout = buf
            print_fn(results, niter, ncall, *args, **kwargs)
            sys.stdout = stdout

            line = buf.getvalue().rstrip()
            print(f"{line} | runtime: {runtime:.1f} min")
            last[0] = now

    return callback


import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import astropy
from astropy.stats import mad_std
from celerite2.backprop import LinAlgError
import numpy as np
from alderaan.constants import *
from alderaan.schema.ephemeris import Ephemeris
from alderaan.schema.litecurve import LiteCurve
from alderaan.schema.planet import Planet
from alderaan.modules.detrend import GaussianProcessDetrender
from alderaan.utils.io import parse_koi_catalog
import warnings

warnings.simplefilter('always', UserWarning)
warnings.filterwarnings(
    action='ignore', category=astropy.units.UnitsWarning, module='astropy'
)

koi_id = 'K00148'

# load KOI catalog
filepath = '/data/user/gjgilbert/projects/alderaan/Catalogs/kepler_dr25_gaia_dr2_crossmatch.csv'
catalog = parse_koi_catalog(filepath, koi_id)

assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

NPL = int(catalog.npl[0])
kic_id = int(catalog.kic_id[0])
planets = [None]*NPL

print(f"{NPL} planets loaded for {koi_id}")

for i in range(NPL):
    planets[i] = Planet(catalog, koi_id, i)
    print(i, planets[i].period)


# load lightcurves
data_dir = '/data/user/gjgilbert/data/MAST_downloads/'
kic_id = catalog.kic_id[0]

litecurve_master_raw = LiteCurve().load_kplr_pdcsap(data_dir, kic_id, 'long cadence')
litecurve_list_raw = litecurve_master_raw.split_quarters()

for i, lc in enumerate(litecurve_list_raw):
    lc = lc.remove_flagged_cadences(bitmask='default')

litecurve_master = LiteCurve().from_list(litecurve_list_raw)

# check for negative timestamps
t_min = litecurve_master.time.min()
t_max = litecurve_master.time.max()

if t_min < 0:
    raise ValueError("Lightcurve has negative timestamps...this will cause problems")

# update planet ephemerides
for n, p in enumerate(planets):
    if p.ephemeris is None:
        _ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
        planets[n] = p.update_ephemeris(_ephemeris)

# split litecurves by quarter
litecurves = litecurve_master.split_quarters()

for j, litecurve in enumerate(litecurves):
    assert len(np.unique(litecurve.quarter)) == 1, "expected one quarter per litecurve"
    assert len(np.unique(litecurve.obsmode)) == 1, "expected one obsmode per litecurve"

print(f"{len(litecurves)} litecurves loaded")


# ################ #
# DETRENDING BLOCK #
# ################ #

# initialize detrenders
detrenders = []
for j, litecurve in enumerate(litecurves):
    detrenders.append(GaussianProcessDetrender(litecurve, planets))

# clip outliers
for j, detrender in enumerate(detrenders):    
    mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
    
    npts_initial = len(detrender.litecurve.time)

    detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=5, mask=mask)
    detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=1000, mask=None)
    
    npts_final = len(detrender.litecurve.time)

    print(f"Quarter {litecurve.quarter[0]} : {npts_initial-npts_final} outliers rejected")

# estimate oscillation periods
oscillation_periods = np.zeros(len(detrenders))
for j, detrender in enumerate(detrenders):
    obsmode = detrender.litecurve.obsmode[0]

    if obsmode == 'short cadence':
        min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
    elif obsmode == 'long cadence':
        min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
    else:
        raise ValueError(f"unsuported obsmode: {obsmode}")

    oscillation_periods[j] = detrender.estimate_oscillation_period(min_period=min_period)

print(np.nanmedian(oscillation_periods))
print(mad_std(oscillation_periods, ignore_nan=True))

# detrend the litecurves
for j, detrender in enumerate(detrenders):
    print(f"Detrending {j+1} of {len(detrenders)} litecurves", flush=True)
    
    # set detrender arguments based on observing mode
    obsmode = detrender.litecurve.obsmode[0]

    if obsmode == 'short cadence':
        min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
        gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_scit * 5 / 2), 91])
        jump_tolerance = 5.0
    elif obsmode == 'long cadence':
        min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
        gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_lcit * 5 / 2), 13])
        jump_tolerance = 5.0
    else:
        raise ValueError(f"unsuported obsmode: {obsmode}")
    
    # make transit mask
    mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
    
    # call detrender.detrend(), using successively simpler models as fallbacks

    try:
        litecurves[j] = detrender.detrend(
            'RotationTerm',
            np.nanmedian(oscillation_periods),
            min_period,
            transit_mask=mask,
            gap_tolerance=gap_tolerance,
            jump_tolerance=jump_tolerance,
            correct_ramp=True,
            return_trend=False, 
            progressbar=False
        )
    except LinAlgError:
        warnings.warn(
            "Initial detrending failed...attempting to refit without exponential ramp component"
        )

        try:
            litecurves[j] = detrender.detrend(
                'RotationTerm',
                np.nanmedian(oscillation_periods),
                min_period,
                transit_mask=mask,
                gap_tolerance=gap_tolerance,
                jump_tolerance=jump_tolerance,
                correct_ramp=False,
                return_trend=False, 
                progressbar=False
            )
        except LinAlgError:
            warnings.warn(
                "Detrending with RotationTerm failed...attempting to detrend with SHOTerm"
            )
            litecurves[j] = detrender.detrend(
                'SHOTerm',
                np.nanmedian(oscillation_periods),
                min_period,
                transit_mask=mask,
                gap_tolerance=gap_tolerance,
                jump_tolerance=jump_tolerance,
                correct_ramp=False,
                return_trend=False, 
                progressbar=False
            )

print("passing")


__all__ = ['LiteCurve']

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from astropy.io import fits
import glob
import lightkurve as lk
import numpy as np
from alderaan.constants import kepler_lcit, kepler_scit


class LiteCurve:
    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            self = self._set_empty_attribute_arrays()
        
        elif len(args) == 1 and isinstance(args[0], list):
            if all([isinstance(lc, LiteCurve) for lc in args[0]]):
                self = self._from_list(*args, **kwargs)
            else:
                raise TypeError("Unexpected input types in list")
        
        elif (len(args) > 1) and isinstance(args[0], str):
            if 'data_source' not in kwargs:
                raise ValueError("Missing required keyword argmument 'data_source")
            else:
                data_source = kwargs.pop('data_source')

            if data_source == 'Kepler PDCSAP':
                self = self._from_kplr_pdcsap(*args, **kwargs)
            elif data_source == 'ALDERAAN':
                self = self._from_alderaan(*args, **kwargs)
            else:
                raise ValueError(f"Unsupported data_source: {data_source}")      
        
        else:
            raise TypeError("Unsupported init signature")


    def _set_empty_attribute_arrays(self):
        self.time = np.array([]).astype(float)
        self.flux = np.array([]).astype(float)
        self.error = np.array([]).astype(float)
        self.cadno = np.array([]).astype(int)
        self.quarter = np.array([]).astype(int)
        self.obsmode = np.array([]).astype(str)
        self.quality = np.array([]).astype(bool)

        return self    


    def _from_list(self, litecurve_list):
        self = self._set_empty_attribute_arrays()

        for i, lc in enumerate(litecurve_list):
            for k in self.__dict__.keys():
                if type(self.__dict__[k]) is np.ndarray:
                    self.__setattr__(k, np.hstack([self.__dict__[k],lc.__dict__[k]]))

        return self
    

    def _from_kplr_pdcsap(self, data_dir, kic_id, obsmode, quarters=None):
        """
        Load photometric data from Kepler Project PDCSAP Flux lightcurves
        The raw fits files must be pre-downloaded from MAST servers and stored locally
        
        This function performs minimal detrending steps
         * remove_nans()
         * normalize()
                
        Arguments:
            data_dir (str) : path to where data are stored
            kic_id (int) : Kepler Input Catalog (KIC) identification number
            obsmode (str) : 'short cadence' or 'long cadence'
            quarters (list) : optional, list of quarters to load
        """
        # sanitize inputs
        if quarters is None:
            quarters = np.arange(18, dtype=int)
        if isinstance(quarters, int):
            quarters = [quarters]

        # load the raw MAST files using lightcurve
        mast_files = glob.glob(data_dir + f"kplr{kic_id:09d}*.fits")
        mast_files.sort()
        
        mast_data_list = []
        for i, mf in enumerate(mast_files):
            with fits.open(mf) as hdu_list:
                if hdu_list[0].header["OBSMODE"] == obsmode and np.isin(
                    hdu_list[0].header["QUARTER"], quarters
                ):
                    mast_data_list.append(lk.read(mf))

        lk_col_raw = lk.LightCurveCollection(mast_data_list)

        # clean up the Collection data structure
        quarters = []
        for lkc in lk_col_raw:
            quarters.append(lkc.quarter)

        lk_col_clean = []
        for q in np.unique(quarters):
            lkc_list = []
            cadno = []

            for lkc in lk_col_raw:
                if (lkc.quarter == q) * (lkc.targetid == kic_id):
                    lkc_list.append(lkc)
                    cadno.append(lkc.cadenceno.min())

            order = np.argsort(cadno)
            lkc_list = [lkc_list[j] for j in order]

            # lk.stitch() also normalizes the lightkurves
            lkc = lk.LightCurveCollection(lkc_list).stitch().remove_nans()
            
            lkc.quarter = lkc.quarter*np.ones(len(lkc.time), dtype='int')
            lkc.season = lkc.quarter % 4
            
            lk_col_clean.append(lkc)

        lk_col_clean = lk.LightCurveCollection(lk_col_clean)

        # stitch into a single LightCurve
        lklc = lk_col_clean.stitch()

        # set LiteCurve attributes
        self.time = np.array(lklc.time.value, dtype=float)
        self.flux = np.array(lklc.flux.value, dtype=float)
        self.error = np.array(lklc.flux_err.value, dtype=float)
        self.cadno = np.array(lklc.cadenceno.value, dtype=int)
        self.quarter = np.array(lklc.quarter, dtype=int)
        self.season = np.array(lklc.season, dtype=int)
        self.obsmode = np.array([obsmode]*len(self.cadno), dtype=str)
        
        # remove cadences flagged by Kepler project pipeline
        self = self._remove_flagged_cadences(lklc.quality)

        return self
        

    def _from_alderaan(self, data_dir, target_id):
        raise NotImplementedError("Loading ALDERAAN files not yet implemented")
    
    
    def split_quarters(self):
        quarters = np.unique(self.quarter)

        litecurve_list = []
        for q in quarters:
            litecurve = LiteCurve()
            for k in litecurve.__dict__.keys():
                if type(litecurve.__dict__[k]) is np.ndarray:
                    litecurve.__setattr__(k, self.__dict__[k][self.quarter == q])
            litecurve_list.append(litecurve)

        return litecurve_list
    

    def _remove_flagged_cadences(self, quality_flags, bitmask='default'):
        qmask = lk.KeplerQualityFlags.create_quality_mask(
            quality_flags, bitmask=bitmask
        )
        for k in self.__dict__.keys():
            if type(self.__dict__[k]) is np.ndarray:
                self.__setattr__(k, self.__dict__[k][qmask])

        self.quality = np.ones(len(self.time), dtype=bool)

        return self


import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import astropy
import numpy as np
from alderaan.schema.litecurve import LiteCurve
import warnings

warnings.simplefilter('always', UserWarning)

# supress UnitsWarnings (this code doesn't use astropy units)
warnings.filterwarnings(
    action='ignore', category=astropy.units.UnitsWarning, module='astropy'
)

data_dir = '/data/user/gjgilbert/data/MAST_downloads/'
kic_id = 5735762

litecurve_raw = LiteCurve().load_kplr_pdcsap(data_dir, kic_id, 'long cadence')
litecurve_list = litecurve_raw.split_quarters()

for i, lc in enumerate(litecurve_list):
    lc = lc.remove_flagged_cadences(bitmask='default')

litecurve_clean = LiteCurve().from_list(litecurve_list)

if np.min(litecurve_clean.time) < 0:
    raise ValueError("Lightcurve has negative timestamps...this will cause problems")

print("passing")


import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from aesara_theano_fallback import aesara as theano
import astropy
from astropy.stats import mad_std
from celerite2.backprop import LinAlgError
from datetime import datetime
import gc
import matplotlib.pyplot as plt
import numpy as np
import shutil
from alderaan.constants import *
from alderaan.schema.ephemeris import Ephemeris
from alderaan.schema.litecurve import LiteCurve
from alderaan.schema.planet import Planet
from alderaan.modules.base import BaseAlg
from alderaan.modules.detrend import GaussianProcessDetrender
from alderaan.modules.omc import OMC
from alderaan.modules.quality_control import QualityControl
from alderaan.modules.quicklook import plot_litecurve, plot_omc
from alderaan.utils.io import parse_koi_catalog, parse_holczer16_catalog
from timeit import default_timer as timer
import warnings

# #################### #
# Initialization Block #
########################

# flush buffer
sys.stdout.flush()
sys.stderr.flush()

# filter warnings
warnings.simplefilter('always', UserWarning)
warnings.filterwarnings(
    action='ignore', category=astropy.units.UnitsWarning, module='astropy'
)

# start program timer
global_start_time = timer()

print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Pipeline")
print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
print("+" * shutil.get_terminal_size().columns)
print("")

# hard-code inputs
mission = 'Kepler'
target = 'K00137'
run_id = 'develop'

project_dir = '/Users/research/projects/alderaan/'
data_dir = '/Users/research/data/MAST_downloads/'
catalog_csv = os.path.join(project_dir, 'Catalogs/kepler_dr25_gaia_dr2_crossmatch.csv')

print("")
print(f"   MISSION : {mission}")
print(f"   TARGET  : {target}")
print(f"   RUN ID  : {run_id}")
print("")
print(f"   Project directory : {project_dir}")
print(f"   Data directory    : {data_dir}")
print(f"   Input catalog     : {catalog_csv}")
print("")
print(f"   theano cache : {theano.config.compiledir}")
print("")

# build directory structure
outputs_dir = os.path.join(project_dir, 'outputs')
os.makedirs(outputs_dir, exist_ok=True)

results_dir = os.path.join(outputs_dir, 'results', run_id, target)
os.makedirs(results_dir, exist_ok=True)

quicklook_dir = os.path.join(outputs_dir, 'quicklook', run_id, target)
os.makedirs(quicklook_dir, exist_ok=True)

# ######### #
# I/O Block #
# ######### #

print('\n\nI/O BLOCK\n')

# load KOI catalog
catalog = parse_koi_catalog(catalog_csv, target)

assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

NPL = int(catalog.npl[0])
koi_id = catalog.koi_id[0]
kic_id = int(catalog.kic_id[0])

# load lightcurves
litecurve_master = LiteCurve(data_dir, kic_id, 'long cadence', data_source='Kepler PDCSAP')

t_min = litecurve_master.time.min()
t_max = litecurve_master.time.max()
if t_min < 0:
    raise ValueError("Lightcurve has negative timestamps...this will cause problems")

# split litecurves by quarter
litecurves = litecurve_master.split_quarters()

for j, litecurve in enumerate(litecurves):
    assert len(np.unique(litecurve.quarter)) == 1, "expected one quarter per litecurve"
    assert len(np.unique(litecurve.obsmode)) == 1, "expected one obsmode per litecurve"

print(f"{len(litecurves)} litecurves loaded for {target}")

# initialize planets (catch no ephemeris warning)
with warnings.catch_warnings(record=True) as catch:
    warnings.simplefilter('always', category=UserWarning)

    planets = [None]*NPL
    for n in range(NPL):
        planets[n] = Planet(catalog, target, n)

print(f"\n{NPL} planets loaded for {target}")
print([np.round(p.period,6) for p in planets])

# update planet ephemerides
for n, p in enumerate(planets):
    if p.ephemeris is None:
        ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
        planets[n] = p.update_ephemeris(ephemeris)

# load Holczer+2016 catalog
filepath = os.path.join(project_dir, 'Catalogs/holczer_2016_kepler_ttvs.txt')
holczer_ephemerides = parse_holczer16_catalog(filepath, koi_id, NPL)

print(f"\n{len(holczer_ephemerides)} ephemerides found in Holczer+2016")

# match Holczer ephemerides to Planets
count = 0

for n, p in enumerate(planets):
    for ephem in holczer_ephemerides:
        match = np.isclose(ephem.period, p.period, rtol=0.01, atol=p.duration)

        if match:
            print(f"  Planet {n} : {p.period:.6f} --> {ephem.period:.6f}")
            planets[n] = p.update_ephemeris(ephem)
            count += 1

print(f"{count} matching ephemerides found ({len(holczer_ephemerides)} expected)")

# update ephemerides
for n, p in enumerate(planets):
    p.ephemeris = p.ephemeris.interpolate('spline', full=True)
    planets[n] = p.update_ephemeris(p.ephemeris)

# quicklook litecurve
filepath = os.path.join(quicklook_dir, f"{target}_litecurve_raw.png")
_ = plot_litecurve(litecurve_master, target, planets, filepath)

# end-of-block cleanup
sys.stdout.flush()
sys.stderr.flush()
plt.close('all')
gc.collect()

print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


# ##################### #
# QUALITY CONTROL BLOCK #
# ##################### #

print("\n\nQUALITY CONTROL BLOCK")

qc = QualityControl(LiteCurve(litecurves), planets)

# check for transits with poor photometric coverage
with warnings.catch_warnings(record=True) as catch:
    warnings.simplefilter('always', category=RuntimeWarning)
    coverage = qc.check_coverage()

# check for transits with unusually high noise
with warnings.catch_warnings(record=True) as catch:
    warnings.simplefilter('always', category=RuntimeWarning)
    good_rms = qc.check_rms(rel_size=3.0, abs_size=2/24, sigma_cut=5.0)

for n, p in enumerate(planets):
    print(f"\n  Planet {n}:")

    assert len(coverage[n]) == len(p.ephemeris.ttime)
    assert len(good_rms[n]) == len(p.ephemeris.ttime)

    planets[n].ephemeris.quality = coverage[n] & good_rms[n]

    _nbad = np.sum(~coverage[n])
    _ntot = len(coverage[n])
    print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for insufficent photometric coverage")

    _nbad = np.sum(~good_rms[n] & coverage[n])
    _ntot = len(good_rms[n])
    print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for high photometric noise")

# end-of-block cleanup
sys.stdout.flush()
sys.stderr.flush()
plt.close('all')
gc.collect()

print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")



import os
import sys

base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))
if base_path not in sys.path:
    sys.path.insert(0, base_path)

from aesara_theano_fallback import aesara as theano
import argparse
import astropy
from astropy.stats import mad_std
from celerite2.backprop import LinAlgError
from configparser import ConfigParser
from datetime import datetime
import gc
import matplotlib.pyplot as plt
import numpy as np
import shutil
from alderaan.constants import *
from alderaan.schema.ephemeris import Ephemeris
from alderaan.schema.litecurve import LiteCurve
from alderaan.schema.planet import Planet
from alderaan.modules.detrend import GaussianProcessDetrender
from alderaan.modules.omc import OMC
from alderaan.modules.transit_model.transit_model import ShapeTransitModel, TTimeTransitModel
from alderaan.modules.quality_control import QualityControl
from alderaan.modules.quicklook import plot_litecurve, plot_omc, dynesty_cornerplot, dynesty_runplot, dynesty_traceplot
from alderaan.utils.io import expand_config_path, parse_koi_catalog, parse_holczer16_catalog, copy_input_target_catalog
from timeit import default_timer as timer
import warnings


def initialize_pipeline():
    # flush buffer
    sys.stdout.flush()
    sys.stderr.flush()

    # filter warnings
    warnings.simplefilter('always', UserWarning)
    warnings.filterwarnings(
        action='ignore', category=astropy.units.UnitsWarning, module='astropy'
    )

    # start timer
    global_start_time = timer()

    return global_start_time


def cleanup():
    sys.stdout.flush()
    sys.stderr.flush()
    plt.close('all')
    gc.collect()


def main():
    # initialize program
    global_start_time = initialize_pipeline()

    print("")
    print("+" * shutil.get_terminal_size().columns)
    print("ALDERAAN Pipeline")
    print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
    print("+" * shutil.get_terminal_size().columns)
    print("")

    # read inputs from config
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--mission', required=True, type=str, help='Observing mission')
    parser.add_argument('-t', '--target', required=True, type=str, help='Target ID for star')
    parser.add_argument('-c', '--config', required=True, type=str, help='Path to config file')
    args = parser.parse_args()

    config = ConfigParser()
    config.read(os.path.join(base_path, args.config))

    mission = args.mission
    target = args.target
    run_id = config['RUN']['run_id']

    data_dir =  expand_config_path(config['PATHS']['data_dir'])
    outputs_dir = expand_config_path(config['PATHS']['outputs_dir'])
    catalog_dir = expand_config_path(config['PATHS']['catalog_dir'])

    catalog_csv = os.path.join(catalog_dir, str(config['ARGS']['catalog_csv']))

    print("")
    print(f"   MISSION : {mission}")
    print(f"   TARGET  : {target}")
    print(f"   RUN ID  : {run_id}")
    print("")
    print(f"   Base path         : {base_path}")
    print(f"   Data directory    : {data_dir}")
    print(f"   Config file       : {args.config}")
    print(f"   Input catalog     : {os.path.basename(catalog_csv)}")
    print("")
    print(f"   theano cache : {theano.config.compiledir}")
    print("")

    # build directory structure
    os.makedirs(outputs_dir, exist_ok=True)

    results_dir = os.path.join(outputs_dir, 'results', run_id, target)
    os.makedirs(results_dir, exist_ok=True)

    quicklook_dir = os.path.join(outputs_dir, 'quicklook', run_id, target)
    os.makedirs(quicklook_dir, exist_ok=True)

    # copy input catalog into results directory
    catalog_csv_copy = os.path.join(outputs_dir, 'results', run_id, f'{run_id}.csv')
    copy_input_target_catalog(catalog_csv, catalog_csv_copy)


    # ######### #
    # I/O Block #
    # ######### #

    print('\n\nI/O BLOCK\n')

    # load KOI catalog
    catalog = parse_koi_catalog(catalog_csv, target)

    assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

    NPL = int(catalog.npl[0])
    koi_id = catalog.koi_id[0]
    kic_id = int(catalog.kic_id[0])

    # load lightcurves
    litecurve_master = LiteCurve(data_dir, kic_id, 'long cadence', data_source='Kepler PDCSAP')

    t_min = litecurve_master.time.min()
    t_max = litecurve_master.time.max()
    if t_min < 0:
        raise ValueError("Lightcurve has negative timestamps...this will cause problems")

    # split litecurves by quarter
    litecurves = litecurve_master.split_quarters()

    for j, litecurve in enumerate(litecurves):
        assert len(np.unique(litecurve.quarter)) == 1, "expected one quarter per litecurve"
        assert len(np.unique(litecurve.obsmode)) == 1, "expected one obsmode per litecurve"

    print(f"{len(litecurves)} litecurves loaded for {target}")

    # initialize planets (catch no ephemeris warning)
    with warnings.catch_warnings(record=True) as catch:
        warnings.simplefilter('always', category=UserWarning)
        planets = [None]*NPL
        for n in range(NPL):
            planets[n] = Planet(catalog, target, n)

    print(f"\n{NPL} planets loaded for {target}")
    print([np.round(p.period,6) for p in planets])

    # update planet ephemerides
    for n, p in enumerate(planets):
        if p.ephemeris is None:
            _ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
            planets[n] = p.update_ephemeris(_ephemeris)

    # load Holczer+2016 catalog
    filepath = os.path.join(catalog_dir, 'holczer_2016_kepler_ttvs.txt')
    holczer_ephemerides = parse_holczer16_catalog(filepath, koi_id, NPL)

    print(f"\n{len(holczer_ephemerides)} ephemerides found in Holczer+2016")

    # match Holczer ephemerides to Planets
    count = 0

    for n, p in enumerate(planets):
        for ephem in holczer_ephemerides:
            match = np.isclose(ephem.period, p.period, rtol=0.01, atol=p.duration)

            if match:
                print(f"  Planet {n} : {p.period:.6f} --> {ephem.period:.6f}")
                planets[n] = p.update_ephemeris(ephem)
                count += 1

    print(f"{count} matching ephemerides found ({len(holczer_ephemerides)} expected)")

    # quicklook litecurve
    filepath = os.path.join(quicklook_dir, f"{target}_litecurve_raw.png")
    _ = plot_litecurve(litecurve_master, target, planets, filepath)

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ######### #
    # OMC Block #
    # ######### #

    print('\n\nOMC BLOCK (initialization)\n')
    print("regularizing ephemerides")

    # initialize OMC object for each planet
    omc_list = []
    for n, p in enumerate(planets):
        omc_list.append(OMC(p.ephemeris))

    # fit a regularized model
    for n, p in enumerate(planets):
        omc = omc_list[n]
        npts = np.sum(omc.quality)

        _period = np.copy(p.period)

        # Matern-3/2 model | don't use GP on very noisy data
        if (npts >= 8) & (np.median(omc.yerr) <= 0.5 * mad_std(omc.yobs)):
            with warnings.catch_warnings(record=True) as catch:
                warnings.simplefilter('always', category=RuntimeWarning)
                trace = omc.sample(omc.matern32_model())

        # Polynomial model | require 2^N transits
        else:
            polyorder = np.max([1, np.min([3, int(np.log2(npts))-1])])
            with warnings.catch_warnings(record=True) as catch:
                warnings.simplefilter('always', category=RuntimeWarning)
                trace = omc.sample(omc.poly_model(polyorder))

        if len(catch) > 0:
            print(f"{len(catch)} RuntimeWarnings caught during sampling")

        # update ephemeris
        omc.ymod = np.nanmedian(trace['pred'], 0)
        omc_list[n] = omc

        p.ephemeris = p.ephemeris.update_from_omc(omc)
        p.ephemeris = p.ephemeris.interpolate('spline', full=True)
        planets[n] = p.update_ephemeris(p.ephemeris)

        # make quicklook plot
        filepath = os.path.join(quicklook_dir, f"{target}_omc_initial.png")
        _ = plot_omc(omc_list, target, filepath)

        print(f"Planet {n} : {_period:.6f} --> {planets[n].period:.6f}")

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ################ #
    # DETRENDING BLOCK #
    # ################ #

    print('\n\nDETRENDING BLOCK (1st pass)\n')

    # initialize detrenders
    detrenders = []
    for j, litecurve in enumerate(litecurves):
        detrenders.append(GaussianProcessDetrender(litecurve, planets))

    # clip outliers
    for j, detrender in enumerate(detrenders):    
        mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
        
        npts_initial = len(detrender.litecurve.time)

        detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=5, mask=mask)
        detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=1000, mask=None)
        
        npts_final = len(detrender.litecurve.time)

        print(f"  Quarter {detrender.litecurve.quarter[0]} : {npts_initial-npts_final} outliers rejected")

    # estimate oscillation periods
    oscillation_periods = np.zeros(len(detrenders))
    for j, detrender in enumerate(detrenders):
        obsmode = detrender.litecurve.obsmode[0]

        if obsmode == 'short cadence':
            min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
        elif obsmode == 'long cadence':
            min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
        else:
            raise ValueError(f"unsuported obsmode: {obsmode}")

        oscillation_periods[j] = detrender.estimate_oscillation_period(min_period=min_period)

    osc_per_mu = np.nanmedian(oscillation_periods)
    osc_per_sd = np.max([mad_std(oscillation_periods, ignore_nan=True), 0.1*osc_per_mu])

    print(f"\nNominal stellar oscillation period = {osc_per_mu:.1f} +/- {osc_per_sd:.1f} days")

    # detrend the litecurves
    for j, detrender in enumerate(detrenders):
        print(f"\nDetrending {j+1} of {len(detrenders)} litecurves", flush=True)
        
        # set detrender arguments based on observing mode
        obsmode = detrender.litecurve.obsmode[0]

        if obsmode == 'short cadence':
            min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
            gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_scit * 5 / 2), 91])
            jump_tolerance = 5.0
        elif obsmode == 'long cadence':
            min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
            gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_lcit * 5 / 2), 13])
            jump_tolerance = 5.0
        else:
            raise ValueError(f"unsuported obsmode: {obsmode}")
        
        # make transit mask
        transit_mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
        
        # call detrender.detrend(), using successively simpler models as fallbacks
        try:
            litecurves[j] = detrender.detrend(
                'RotationTerm',
                np.nanmedian(oscillation_periods),
                min_period,
                transit_mask=transit_mask,
                gap_tolerance=gap_tolerance,
                jump_tolerance=jump_tolerance,
                correct_ramp=True,
                return_trend=False, 
                progressbar=False
            )
        except LinAlgError:
            warnings.warn(
                "Initial detrending failed...attempting to refit without exponential ramp component"
            )

            try:
                litecurves[j] = detrender.detrend(
                    'RotationTerm',
                    np.nanmedian(oscillation_periods),
                    min_period,
                    transit_mask=transit_mask,
                    gap_tolerance=gap_tolerance,
                    jump_tolerance=jump_tolerance,
                    correct_ramp=False,
                    return_trend=False, 
                    progressbar=False
                )
            except LinAlgError:
                warnings.warn(
                    "Detrending with RotationTerm failed...attempting to detrend with SHOTerm"
                )
                litecurves[j] = detrender.detrend(
                    'SHOTerm',
                    np.nanmedian(oscillation_periods),
                    min_period,
                    transit_mask=transit_mask,
                    gap_tolerance=gap_tolerance,
                    jump_tolerance=jump_tolerance,
                    correct_ramp=False,
                    return_trend=False, 
                    progressbar=False
                )

    # recombine litecurves
    litecurve = LiteCurve(litecurves)

    # quicklook litecurve
    filepath = os.path.join(quicklook_dir, f"{target}_litecurve_detrended.png")
    _ = plot_litecurve(litecurve, target, planets, filepath)

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ##################### #
    # QUALITY CONTROL BLOCK #
    # ##################### #

    print("\n\nQUALITY CONTROL BLOCK")

    qc = QualityControl(litecurve, planets)

    # check for transits with poor photometric coverage
    with warnings.catch_warnings(record=True) as catch:
        warnings.simplefilter('always', category=RuntimeWarning)
        coverage = qc.check_coverage()

    # check for transits with unusually high noise
    with warnings.catch_warnings(record=True) as catch:
        warnings.simplefilter('always', category=RuntimeWarning)
        good_rms = qc.check_rms(rel_size=3.0, abs_size=2/24, sigma_cut=5.0)

    for n, p in enumerate(planets):
        print(f"\n  Planet {n}:")

        assert len(coverage[n]) == len(p.ephemeris.ttime)
        assert len(good_rms[n]) == len(p.ephemeris.ttime)

        _nbad = np.sum(~coverage[n])
        _ntot = len(coverage[n])
        print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for insufficent photometric coverage")

        _nbad = np.sum(~good_rms[n] & coverage[n])
        _ntot = len(good_rms[n])
        print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for high photometric noise")
        
        planets[n].ephemeris.quality = coverage[n] & good_rms[n]

    # end-of-block cleanup
    cleanup()

    print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


    # ################### #
    # TRANSIT MODEL BLOCK #
    # ################### #

    print('\n\nTRANSIT MODEL BLOCK\n')

    limbdark = [catalog.limbdark_1[0], catalog.limbdark_2[0]]
    transitmodel = ShapeTransitModel(litecurve, planets, limbdark)

    print("Supersample factor")
    for obsmode in transitmodel.unique_obsmodes:
        print(f"  {obsmode} : {transitmodel._obsmode_to_supersample(obsmode)}")

    print("\nFitting initial transit model")
    theta = transitmodel.optimize()
    planets = transitmodel.update_planet_parameters(theta)
    limbdark = transitmodel.update_limbdark_parameters(theta)

    print("\nFitting independent transit times")
    ttvmodel = TTimeTransitModel(litecurve, planets, limbdark)

    for n, p in enumerate(planets):
        print(f"\nPlanet {n} : fitting {np.sum(p.ephemeris.quality)} transit times")

        ttime_new, ttime_err_new = ttvmodel.mazeh13_holczer16_method(n, quicklook_dir=quicklook_dir)

        assert len(ttime_new) == len(ttime_err_new)
        assert len(ttime_new) == len(p.ephemeris.ttime)

        _nfit = np.sum(~np.isnan(ttime_new))
        _ntot = np.sum(p.ephemeris.quality)

        print(f"Planet {n} : {_nfit} of {_ntot} transit times ({_nfit / _ntot * 100:.1f}%) fit successfully")

    print("\nSampling with DynamicNestedSampler")
    results = transitmodel.sample(progress_every=10)

    filepath = os.path.join(quicklook_dir, f'{koi_id}_dynesty_runplot.png')
    fig, ax = dynesty_runplot(results, koi_id, filepath=filepath)

    for n, p in enumerate(transitmodel.planets):
        filepath = os.path.join(quicklook_dir, f'{koi_id}_dynesty_traceplot_{n:02d}.png')
        fig, ax = dynesty_traceplot(results, koi_id, n, filepath=filepath)

        filepath=os.path.join(quicklook_dir, f'{koi_id}_dynesty_cornerplot_{n:02d}.png')
        fig, ax = dynesty_cornerplot(results, koi_id, n, filepath=filepath, interactive=True)


if __name__ == '__main__':
    main()


import os
import sys

base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
if base_path not in sys.path:
    sys.path.insert(0, base_path)

from aesara_theano_fallback import aesara as theano
import astropy
from astropy.stats import mad_std
from celerite2.backprop import LinAlgError
from datetime import datetime
import gc
import matplotlib.pyplot as plt
import numpy as np
import shutil
from alderaan.constants import *
from alderaan.schema.planet import Planet
from alderaan.schema.ephemeris import Ephemeris
from alderaan.schema.litecurve import LiteCurve
from alderaan.modules.detrend import GaussianProcessDetrender
from alderaan.modules.transit_model.transit_model import ShapeTransitModel, TTimeTransitModel
from alderaan.modules.quality_control import QualityControl
from alderaan.modules.quicklook import plot_litecurve, dynesty_cornerplot, dynesty_runplot, dynesty_traceplot
from alderaan.utils.io import expand_config_path, parse_koi_catalog, parse_holczer16_catalog
from timeit import default_timer as timer
import warnings

# #################### #
# Initialization Block #
########################

# flush buffer
sys.stdout.flush()
sys.stderr.flush()

# filter warnings
warnings.simplefilter('always', UserWarning)
warnings.filterwarnings(
    action='ignore', category=astropy.units.UnitsWarning, module='astropy'
)

# start program timer
global_start_time = timer()

print("")
print("+" * shutil.get_terminal_size().columns)
print("ALDERAAN Pipeline")
print(f"Initialized {datetime.now().strftime('%d-%b-%Y at %H:%M:%S')}")
print("+" * shutil.get_terminal_size().columns)
print("")

# read inputs from config
from configparser import ConfigParser

config = ConfigParser()
config.read(os.path.join(base_path, 'configs/default_config.cfg'))

mission = config['TARGET']['mission']
target = config['TARGET']['target']
run_id = config['RUN']['run_id']

data_dir =  expand_config_path(config['PATHS']['data_dir'])
outputs_dir = expand_config_path(config['PATHS']['outputs_dir'])
catalog_dir = expand_config_path(config['PATHS']['catalog_dir'])

catalog_csv = os.path.join(catalog_dir, str(config['PATHS']['catalog_csv']))

print("")
print(f"   MISSION : {mission}")
print(f"   TARGET  : {target}")
print(f"   RUN ID  : {run_id}")
print("")
print(f"   Base path         : {base_path}")
print(f"   Data directory    : {data_dir}")
print(f"   Input catalog     : {os.path.basename(catalog_csv)}")
print("")
print(f"   theano cache : {theano.config.compiledir}")
print("")

# build directory structure
os.makedirs(outputs_dir, exist_ok=True)

results_dir = os.path.join(outputs_dir, 'results', run_id, target)
os.makedirs(results_dir, exist_ok=True)

quicklook_dir = os.path.join(outputs_dir, 'quicklook', run_id, target)
os.makedirs(quicklook_dir, exist_ok=True)

# copy input catalog into results directory
catalog_csv_copy = os.path.join(outputs_dir, 'results', run_id, f'{run_id}.csv')

# ######### #
# I/O Block #
# ######### #

print('\n\nI/O BLOCK\n')

# load KOI catalog
catalog = parse_koi_catalog(catalog_csv, target)

assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

NPL = int(catalog.npl[0])
koi_id = catalog.koi_id[0]
kic_id = int(catalog.kic_id[0])

# load lightcurves
litecurve_master = LiteCurve(data_dir, kic_id, 'long cadence', data_source='Kepler PDCSAP', quarters=[2])

t_min = litecurve_master.time.min()
t_max = litecurve_master.time.max()
if t_min < 0:
    raise ValueError("Lightcurve has negative timestamps...this will cause problems")

# split litecurves by quarter
litecurves = litecurve_master.split_quarters()

for j, litecurve in enumerate(litecurves):
    assert len(np.unique(litecurve.quarter)) == 1, "expected one quarter per litecurve"
    assert len(np.unique(litecurve.obsmode)) == 1, "expected one obsmode per litecurve"

print(f"{len(litecurves)} litecurves loaded for {target}")

# initialize planets (catch no ephemeris warning)
with warnings.catch_warnings(record=True) as catch:
    warnings.simplefilter('always', category=UserWarning)

    planets = [None]*NPL
    for n in range(NPL):
        planets[n] = Planet(catalog, target, n)

print(f"\n{NPL} planets loaded for {target}")
print([np.round(p.period,6) for p in planets])

# update planet ephemerides
for n, p in enumerate(planets):
    if p.ephemeris is None:
        ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
        planets[n] = p.update_ephemeris(ephemeris)

# load Holczer+2016 catalog
filepath = os.path.join(catalog_dir, 'holczer_2016_kepler_ttvs.txt')
holczer_ephemerides = parse_holczer16_catalog(filepath, koi_id, NPL)

print(f"\n{len(holczer_ephemerides)} ephemerides found in Holczer+2016")

# match Holczer ephemerides to Planets
count = 0

for n, p in enumerate(planets):
    for ephem in holczer_ephemerides:
        match = np.isclose(ephem.period, p.period, rtol=0.01, atol=p.duration)

        if match:
            print(f"  Planet {n} : {p.period:.6f} --> {ephem.period:.6f}")
            planets[n] = p.update_ephemeris(ephem)
            count += 1

print(f"{count} matching ephemerides found ({len(holczer_ephemerides)} expected)")

# update ephemerides
for n, p in enumerate(planets):
    p.ephemeris = p.ephemeris.interpolate('spline', full=True)
    planets[n] = p.update_ephemeris(p.ephemeris)

# end-of-block cleanup
sys.stdout.flush()
sys.stderr.flush()
plt.close('all')
gc.collect()

print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


# ################ #
# DETRENDING BLOCK #
# ################ #

print('\n\nDETRENDING BLOCK (1st pass)\n')

# initialize detrenders
detrenders = []
for j, litecurve in enumerate(litecurves):
    detrenders.append(GaussianProcessDetrender(litecurve, planets))

# clip outliers
for j, detrender in enumerate(detrenders):    
    mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
    
    npts_initial = len(detrender.litecurve.time)

    detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=5, mask=mask)
    detrender.clip_outliers(kernel_size=13, sigma_upper=5, sigma_lower=1000, mask=None)
    
    npts_final = len(detrender.litecurve.time)

    print(f"  Quarter {detrender.litecurve.quarter[0]} : {npts_initial-npts_final} outliers rejected")

# estimate oscillation periods
oscillation_periods = np.zeros(len(detrenders))
for j, detrender in enumerate(detrenders):
    obsmode = detrender.litecurve.obsmode[0]

    if obsmode == 'short cadence':
        min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
    elif obsmode == 'long cadence':
        min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
    else:
        raise ValueError(f"unsuported obsmode: {obsmode}")

    oscillation_periods[j] = detrender.estimate_oscillation_period(min_period=min_period)

osc_per_mu = np.nanmedian(oscillation_periods)
osc_per_sd = np.max([mad_std(oscillation_periods, ignore_nan=True), 0.1*osc_per_mu])

print(f"\nNominal stellar oscillation period = {osc_per_mu:.1f} +/- {osc_per_sd:.1f} days")

# detrend the litecurves
for j, detrender in enumerate(detrenders):
    print(f"\nDetrending {j+1} of {len(detrenders)} litecurves", flush=True)
    
    # set detrender arguments based on observing mode
    obsmode = detrender.litecurve.obsmode[0]

    if obsmode == 'short cadence':
        min_period = np.max([5 * np.max(detrender.durs), 91 * kepler_scit])
        gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_scit * 5 / 2), 91])
        jump_tolerance = 5.0
    elif obsmode == 'long cadence':
        min_period = np.max([5 * np.max(detrender.durs), 13 * kepler_lcit])
        gap_tolerance = np.max([int(np.min(detrender.durs) / kepler_lcit * 5 / 2), 13])
        jump_tolerance = 5.0
    else:
        raise ValueError(f"unsuported obsmode: {obsmode}")
    
    # make transit mask
    transit_mask = detrender.make_transit_mask(rel_size=3.0, abs_size=2/24, mask_type='condensed')
    
    # call detrender.detrend(), using successively simpler models as fallbacks
    try:
        litecurves[j] = detrender.detrend(
            'RotationTerm',
            np.nanmedian(oscillation_periods),
            min_period,
            transit_mask=transit_mask,
            gap_tolerance=gap_tolerance,
            jump_tolerance=jump_tolerance,
            correct_ramp=True,
            return_trend=False, 
            progressbar=False
        )
    except LinAlgError:
        warnings.warn(
            "Initial detrending failed...attempting to refit without exponential ramp component"
        )

        try:
            litecurves[j] = detrender.detrend(
                'RotationTerm',
                np.nanmedian(oscillation_periods),
                min_period,
                transit_mask=transit_mask,
                gap_tolerance=gap_tolerance,
                jump_tolerance=jump_tolerance,
                correct_ramp=False,
                return_trend=False, 
                progressbar=False
            )
        except LinAlgError:
            warnings.warn(
                "Detrending with RotationTerm failed...attempting to detrend with SHOTerm"
            )
            litecurves[j] = detrender.detrend(
                'SHOTerm',
                np.nanmedian(oscillation_periods),
                min_period,
                transit_mask=transit_mask,
                gap_tolerance=gap_tolerance,
                jump_tolerance=jump_tolerance,
                correct_ramp=False,
                return_trend=False, 
                progressbar=False
            )

# recombine litecurves
litecurve = LiteCurve(litecurves)

# quicklook litecurve
filepath = os.path.join(quicklook_dir, f"{target}_litecurve_detrended.png")
_ = plot_litecurve(litecurve, target, planets, filepath)

# end-of-block cleanup
sys.stdout.flush()
sys.stderr.flush()
plt.close('all')
gc.collect()

print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


# ##################### #
# QUALITY CONTROL BLOCK #
# ##################### #

print("\n\nQUALITY CONTROL BLOCK")

qc = QualityControl(litecurve, planets)

# check for transits with poor photometric coverage
with warnings.catch_warnings(record=True) as catch:
    warnings.simplefilter('always', category=RuntimeWarning)
    coverage = qc.check_coverage()

# check for transits with unusually high noise
with warnings.catch_warnings(record=True) as catch:
    warnings.simplefilter('always', category=RuntimeWarning)
    good_rms = qc.check_rms(rel_size=3.0, abs_size=2/24, sigma_cut=5.0)

for n, p in enumerate(planets):
    print(f"\n  Planet {n}:")

    assert len(coverage[n]) == len(p.ephemeris.ttime)
    assert len(good_rms[n]) == len(p.ephemeris.ttime)

    _nbad = np.sum(~coverage[n])
    _ntot = len(coverage[n])
    print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for insufficent photometric coverage")

    _nbad = np.sum(~good_rms[n] & coverage[n])
    _ntot = len(good_rms[n])
    print(f"    {np.sum(_nbad)} of {_ntot} transits ({int(100*_nbad/_ntot)}%) rejected for high photometric noise")
    
    planets[n].ephemeris.quality = coverage[n] & good_rms[n]

# end-of-block cleanup
sys.stdout.flush()
sys.stderr.flush()
plt.close('all')
gc.collect()

print(f"\ncumulative runtime = {((timer()-global_start_time)/60):.1f} min")


# ################### #
# TRANSIT MODEL BLOCK #
# ################### #

print('\n\nTRANSIT MODEL BLOCK\n')

limbdark = [catalog.limbdark_1[0], catalog.limbdark_2[0]]
transitmodel = ShapeTransitModel(litecurve, planets, limbdark)

print("Supersample factor")
for obsmode in transitmodel.unique_obsmodes:
    print(f"  {obsmode} : {transitmodel._obsmode_to_supersample(obsmode)}")

print("\nFitting initial transit model")
theta = transitmodel.optimize()
planets = transitmodel.update_planet_parameters(theta)
limbdark = transitmodel.update_limbdark_parameters(theta)

print("\nFitting independent transit times")
ttvmodel = TTimeTransitModel(litecurve, planets, limbdark)

for n, p in enumerate(planets):
    print(f"\nPlanet {n} : fitting {np.sum(p.ephemeris.quality)} transit times")

    ttime_new, ttime_err_new = ttvmodel.mazeh13_holczer16_method(n, quicklook_dir=quicklook_dir)

    assert len(ttime_new) == len(ttime_err_new)
    assert len(ttime_new) == len(p.ephemeris.ttime)

    _nfit = np.sum(~np.isnan(ttime_new))
    _ntot = np.sum(p.ephemeris.quality)

    print(f"Planet {n} : {_nfit} of {_ntot} transit times ({_nfit / _ntot * 100:.1f}%) fit successfully")

print("\nSampling with DynamicNestedSampler")
results = transitmodel.sample(progress_every=10)

filepath = os.path.join(quicklook_dir, f'{koi_id}_dynesty_runplot.png')
fig, ax = dynesty_runplot(results, koi_id, filepath=filepath)

for n, p in enumerate(transitmodel.planets):
    filepath = os.path.join(quicklook_dir, f'{koi_id}_dynesty_traceplot_{n:02d}.png')
    fig, ax = dynesty_traceplot(results, koi_id, n, filepath=filepath)

    filepath=os.path.join(quicklook_dir, f'{koi_id}_dynesty_cornerplot_{n:02d}.png')
    fig, ax = dynesty_cornerplot(results, koi_id, n, filepath=filepath, interactive=True)

print("passing")


import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import numpy as np
from alderaan.schema.ephemeris import Ephemeris
from alderaan.schema.planet import Planet
from alderaan.utils.io import parse_koi_catalog
import warnings

warnings.simplefilter('always', UserWarning)

filepath = '/data/user/gjgilbert/projects/alderaan/Catalogs/kepler_dr25_gaia_dr2_crossmatch.csv'
koi_id = 'K00148'
catalog = parse_koi_catalog(filepath, koi_id)

assert np.all(np.diff(catalog.period) > 0), "Planets should be ordered by ascending period"

t_min = 0.
t_max = 1400.

NPL = int(catalog.npl[0])
planets = [None]*NPL

for n in range(NPL):
    planets[n] = Planet(catalog, koi_id, n)

for n, p in enumerate(planets):
    ephemeris = Ephemeris(period=p.period, epoch=p.epoch, t_min=t_min, t_max=t_max)
    p = p.update_ephemeris(ephemeris)

print("passing")